Index: sys/dev/iwn/if_iwn.c
===================================================================
--- sys/dev/iwn/if_iwn.c	(revision 234975)
+++ sys/dev/iwn/if_iwn.c	(working copy)
@@ -2758,13 +2758,13 @@ static void
 iwn_ampdu_tx_done(struct iwn_softc *sc, int qid, int idx, int nframes,
     void *stat)
 {
+	struct iwn_ops *ops = &sc->ops;
 	struct ifnet *ifp = sc->sc_ifp;
 	struct iwn_tx_ring *ring = &sc->txq[qid];
 	struct iwn_tx_data *data;
 	struct mbuf *m;
 	struct iwn_node *wn;
 	struct ieee80211_node *ni;
-	struct ieee80211vap *vap;
 	struct ieee80211_tx_ampdu *tap;
 	uint64_t bitmap;
 	uint32_t *status = stat;
@@ -2802,29 +2802,28 @@ iwn_ampdu_tx_done(struct iwn_softc *sc, int qid, i
 		bitmap = bitmap << shift;
 		bitmap |= 1ULL << bit;
 	}
+
 	tap = sc->qid2tap[qid];
-	if (tap != NULL) {
-		tid = WME_AC_TO_TID(tap->txa_ac);
-		wn = (void *)tap->txa_ni;
-		wn->agg[tid].bitmap = bitmap;
-		wn->agg[tid].startidx = start;
-		wn->agg[tid].nframes = nframes;
-	}
+	tid = WME_AC_TO_TID(tap->txa_ac);
+	wn = (void *)tap->txa_ni;
+	wn->agg[tid].bitmap = bitmap;
+	wn->agg[tid].startidx = start;
+	wn->agg[tid].nframes = nframes;
 
 	seqno = le32toh(*(status + nframes)) & 0xfff;
 	for (lastidx = (seqno & 0xff); ring->read != lastidx;) {
 		data = &ring->data[ring->read];
 
-		KASSERT(data->ni != NULL, ("no node"));
-
 		/* Unmap and free mbuf. */
 		bus_dmamap_sync(ring->data_dmat, data->map,
 		    BUS_DMASYNC_POSTWRITE);
 		bus_dmamap_unload(ring->data_dmat, data->map);
 		m = data->m, data->m = NULL;
 		ni = data->ni, data->ni = NULL;
-		vap = ni->ni_vap;
 
+		KASSERT(ni != NULL, ("no node"));
+		KASSERT(m != NULL, ("no mbuf"));
+
 		if (m->m_flags & M_TXCB)
 			ieee80211_process_callback(ni, m, 1);
 
@@ -2835,6 +2834,18 @@ iwn_ampdu_tx_done(struct iwn_softc *sc, int qid, i
 		ring->read = (ring->read + 1) % IWN_TX_RING_COUNT;
 	}
 
+	if (ring->queued == 0 && IEEE80211_AMPDU_RUNNING(tap) == 0) {
+		DPRINTF(sc, IWN_DEBUG_XMIT, "%s: ra=%d tid=%d ssn=%d qid=%d\n",
+		    __func__, wn->id, tid, tap->txa_start, qid);
+		if (iwn_nic_lock(sc) != 0)
+			return;
+		ops->ampdu_tx_stop(sc, qid, tid, tap->txa_start & 0xfff);
+		iwn_nic_unlock(sc);
+		sc->qid2tap[qid] = NULL;
+		free(tap->txa_private, M_DEVBUF);
+		tap->txa_private = NULL;
+	}
+
 	sc->sc_tx_timer = 0;
 	if (ring->queued < IWN_TX_RING_LOMARK) {
 		sc->qfullmsk &= ~(1 << ring->qid);
@@ -5633,6 +5644,8 @@ iwn_ampdu_tx_start(struct ieee80211com *ic, struct
 	if ((error = iwn_nic_lock(sc)) != 0)
 		return 0;
 	qid = *(int *)tap->txa_private;
+	DPRINTF(sc, IWN_DEBUG_XMIT, "%s: ra=%d tid=%d ssn=%d qid=%d\n",
+	    __func__, wn->id, tid, tap->txa_start, qid);
 	ops->ampdu_tx_start(sc, ni, qid, tid, tap->txa_start & 0xfff);
 	iwn_nic_unlock(sc);
 
@@ -5645,21 +5658,27 @@ iwn_ampdu_tx_stop(struct ieee80211_node *ni, struc
 {
 	struct iwn_softc *sc = ni->ni_ic->ic_ifp->if_softc;
 	struct iwn_ops *ops = &sc->ops;
+	struct iwn_node *wn = (void *)ni;
 	uint8_t tid = WME_AC_TO_TID(tap->txa_ac);
 	int qid;
 
+	sc->sc_addba_stop(ni, tap);
+
 	if (tap->txa_private == NULL)
 		return;
 
 	qid = *(int *)tap->txa_private;
-	if (iwn_nic_lock(sc) != 0)
-		return;
-	ops->ampdu_tx_stop(sc, qid, tid, tap->txa_start & 0xfff);
-	iwn_nic_unlock(sc);
-	sc->qid2tap[qid] = NULL;
-	free(tap->txa_private, M_DEVBUF);
-	tap->txa_private = NULL;
-	sc->sc_addba_stop(ni, tap);
+	if (sc->txq[qid].queued == 0) {
+		DPRINTF(sc, IWN_DEBUG_XMIT, "%s: ra=%d tid=%d ssn=%d qid=%d\n",
+		    __func__, wn->id, tid, tap->txa_start, qid);
+		if (iwn_nic_lock(sc) != 0)
+			return;
+		ops->ampdu_tx_stop(sc, qid, tid, tap->txa_start & 0xfff);
+		iwn_nic_unlock(sc);
+		sc->qid2tap[qid] = NULL;
+		free(tap->txa_private, M_DEVBUF);
+		tap->txa_private = NULL;
+	}
 }
 
 static void